/*
 * Copyright (C) 2021 Anton Romanov - All Rights Reserved
 * You may use, distribute and modify this code, please write to: romanov73@gmail.com.
 */

package ru.ulstu.extractor.heuristic.service;

import com.github.javaparser.JavaParser;
import com.github.javaparser.ParseResult;
import com.github.javaparser.ast.CompilationUnit;
import org.springframework.stereotype.Service;
import ru.ulstu.extractor.heuristic.api.StructuralUnitIdentifier;
import ru.ulstu.extractor.heuristic.component.BuildTool;
import ru.ulstu.extractor.heuristic.component.JavaProgrammingLanguage;
import ru.ulstu.extractor.heuristic.component.ProgrammingLanguage;
import ru.ulstu.extractor.heuristic.model.EntityUnit;
import ru.ulstu.extractor.heuristic.model.StructuralUnit;
import ru.ulstu.extractor.util.StringUtils;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

import static ru.ulstu.extractor.heuristic.service.DetectorService.LangDetectScrupulousness.LOW;

@Service
public class JavaIdentifier extends StructuralUnitIdentifier {
    private final DetectorService detectorService;
    private final BuildToolService buildToolService;
    private final ProgrammingLanguageService programmingLanguageService;

    public JavaIdentifier(DetectorService detectorService,
                          BuildToolService buildToolService,
                          ProgrammingLanguageService programmingLanguageService) {
        this.detectorService = detectorService;
        this.buildToolService = buildToolService;
        this.programmingLanguageService = programmingLanguageService;
    }

    @Override
    public boolean canAppliedToProject(String projectPath, List<File> projectFiles, List<File> rootProjectFiles) {
        return /*getBuildTool() instanceof GradleBuildTool
                &&*/ getMainProgrammingLanguage(projectPath, projectFiles, rootProjectFiles).orElse(null) instanceof JavaProgrammingLanguage;
    }

    public boolean canAppliedToFile(File projectFile) {
        return getMainProgrammingLanguage(projectFile, LOW).orElse(null) instanceof JavaProgrammingLanguage;
    }

    @Override
    public List<StructuralUnit> getEntityClasses(String projectPath, String subDirectory, List<File> projectFiles) {
        return projectFiles.stream()
                .filter(file -> StringUtils.fileInSubdirectory(file.getPath(), projectPath, subDirectory))
                .filter(this::isEntityClass)
                .map(file -> new EntityUnit(projectPath, file))
                .collect(Collectors.toList());
    }

    @Override
    public List<StructuralUnit> getBusinessLogicClasses() {
        return null;
    }

    @Override
    public boolean isMultiModuleProject() {
        return false;
    }

    @Override
    public Optional<BuildTool> getBuildTool(List<File> rootDirectoryFiles) {
        return buildToolService.getProjectBuildTool(rootDirectoryFiles);
    }

    @Override
    protected DetectorService getDetectorService() {
        return detectorService;
    }

    @Override
    protected ProgrammingLanguageService getProgrammingLanguageService() {
        return programmingLanguageService;
    }

    @Override
    protected ProgrammingLanguage getProgrammingLanguage() {
        return new JavaProgrammingLanguage();
    }

    @Override
    public boolean isEntityClass(File file) {
        try {
            return file.getName().endsWith("java") && classContainsAnnotation(file, "@Entity");
        } catch (Exception ex) {
            ex.printStackTrace();
        }
        return false;
    }

    private boolean classContainsAnnotation(File file, String annotationDeclaration) throws FileNotFoundException {
        JavaParser parser = new JavaParser();
        ParseResult<CompilationUnit> parseResult = parser.parse(file);
        if (parseResult.getResult().isPresent() && parseResult.getResult().get().findCompilationUnit().isPresent()) {
            return parseResult.getResult().get().getTypes().stream()
                    .anyMatch(clazz -> clazz.getAnnotations().stream().anyMatch(annotation -> annotation.toString().startsWith(annotationDeclaration)));
        }
        return false;
    }

    @Override
    protected String getSourceDirectory(List<File> rootDirectoryFiles) {
        return buildToolService.getProjectBuildTool(rootDirectoryFiles)
                .map(BuildTool::getSourceDirectoryPath)
                .orElse("src");
    }

    @Override
    protected boolean isBusinessLogicClass(File file) {
        return false;
    }
}
